#ifndef MD_H
#define MD_H

#include <libusb-1.0/libusb.h>
#include <stdarg.h>


//Uncomment to enable debugging
// debugging macros 
/*
#define WHERESTR  "[file %s, %s():, line %d]: "
#define WHEREARG  __FILE__, __func__, __LINE__
#define DBG2(...)       fprintf(stderr, __VA_ARGS__)
#define DBG(_fmt, ...)  DBG2(WHERESTR _fmt, WHEREARG, __VA_ARGS__)
#define DBGPLAIN(...) DBG2(__VA_ARGS__)
//...
*/

//Comment out the next three lines if debugging is enabled
#define DBG2(...) 
#define DBG(_fmt, ...) 
#define DBGPLAIN(...) 

// NetMD Protocol return status (first byte of request)
#define STATUS_CONTROL 0x00
#define STATUS_STATUS 0x01
#define STATUS_SPECIFIC_INQUIRY 0x02
#define STATUS_NOTIFY 0x03
#define STATUS_GENERAL_INQUIRY 0x04
// ... (first byte of response)
#define STATUS_NOT_IMPLEMENTED 0x08
#define STATUS_ACCEPTED 0x09
#define STATUS_REJECTED 0x0a
#define STATUS_IN_TRANSITION 0x0b
#define STATUS_IMPLEMENTED 0x0c
#define STATUS_CHANGED 0x0d
#define STATUS_INTERIM 0x0f


#define ENCODING_SP  0x90
#define ENCODING_LP2  0x92
#define ENCODING_LP4  0x93

#define CHANNELS_MONO  0x01
#define CHANNELS_STEREO  0x00

#define TRACK_FLAG_PROTECTED  0x03

#define DISC_FLAG_WRITABLE  0x10
#define DISC_FLAG_WRITE_PROTECTED  0x40

#define MD_SERIALNO_LENGTH  256

//extern unsigned char serial[];

typedef struct mddevice MDdevice;
struct mddevice{
	int idVendor;
	int idProduct;
	char *description;
	unsigned char *serialno;
	libusb_device_handle *handle;
};

typedef struct dev_id Devid;
struct dev_id{
	int idVendor;
	int idProduct;
	char *description;
};

extern Devid known_dev[];

/*
extern struct dev_id known_devxxxx[29] = {
 { 0x04dd, 0x7202, "Sharp IM-MT899H"},
 {0x054c, 0x0075, "Sony MZ-N1"}, 
 {0x054c, 0x0080, "Sony LAM-1"}, 
 {0x054c, 0x0081, "Sony MDS-JB980"}, 
 {0x054c, 0x0084, "Sony MZ-N505"}, 
 {0x054c, 0x0085, "Sony MZ-S1"}, 
 {0x054c, 0x0086, "Sony MZ-N707"}, 
 {0x054c, 0x00c6, "Sony MZ-N10"}, 
 {0x054c, 0x00c7, "Sony MZ-N910"},
 {0x054c, 0x00c8, "Sony MZ-N710/NF810"}, 
 {0x054c, 0x00c9, "Sony MZ-N510/N610"}, 
 {0x054c, 0x00ca, "Sony MZ-NE410/NF520D"}, 
 {0x054c, 0x00eb, "Sony MZ-NE810/NE910"},
 {0x054c, 0x0101, "Sony LAM-10"},
 {0x054c, 0x0113, "Aiwa AM-NX1"},
 {0x054c, 0x014c, "Aiwa AM-NX9"},
 {0x054c, 0x017e, "Sony MZ-NH1"},
 {0x054c, 0x0180, "Sony MZ-NH3D"},
 {0x054c, 0x0182, "Sony MZ-NH900"},
 {0x054c, 0x0184, "Sony MZ-NH700/NH800"},
 {0x054c, 0x0186, "Sony MZ-NH600/NH600D"},
 {0x054c, 0x0188, "Sony MZ-N920"},
 {0x054c, 0x018a, "Sony LAM-3"},
 {0x054c, 0x01e9, "Sony MZ-DH10P"},
 {0x054c, 0x0219, "Sony MZ-RH10"},
 {0x054c, 0x021b, "Sony MZ-RH710/MZ-RH910"},
 {0x054c, 0x022c, "Sony CMT-AH10 (stereo set with integrated MD)"},
 {0x054c, 0x023c, "Sony DS-HMD1 (device without analog music rec/playback)"},
 {0x054c, 0x0286, "Sony MZ-RH1"}
};
*/



typedef struct trk_node Tracknode;
struct trk_node{
	Tracknode *next_trk;
	unsigned char channels; //mono || stereo
	unsigned char encoding;	// SP / LP2 / LP4
	//unsigned char *protected;
	unsigned char gaoled;
	unsigned char *title_ascii;
	unsigned char *title_wchar;
	int track_no;
	int *length;
	int r_date;
	int r_time;
};

typedef struct grp_node Groupnode;
struct grp_node{
	char *name;
	int nos_tracks;
	Tracknode *tlist;	
	Groupnode *next_group;
};

struct disc{
	int *capacity; //Divide into 3x4??
	int nos_tracks;
	int nos_groups;
	unsigned char *title;
	unsigned char disc_flags;
	Groupnode *groups; //Tracks that do not belong to a group are placed in a group and listed as tracks that do not belong to a group
};

typedef struct tracklist{
	Tracknode *t;
	Tracknode *nextt;
} Tracklist;

typedef struct tglist{
	int nos_groups; //move to disc struct perhaps?
	Groupnode *g;
} Tglist;



//Tglist *makeTGlist();
Groupnode *makeTGlist();
Tracknode *makeTrackList(int, int);
Tracknode *makeArrayTrackList(int *, int);
void addGroup(Groupnode **, Groupnode *);
void addTrack(Tracknode **, Tracknode *);

unsigned getbits(unsigned, int, int);
int querySize(unsigned char *, va_list ); //va_list
unsigned char* formatQuery(int *, unsigned char *, ...);
unsigned char* scanQuery(int, unsigned char *, unsigned char *);
int mdInit();
int mdClose();
unsigned char getDiscFlags();
int dumpResultToFile(unsigned char *, int);
int getTrackCount();
unsigned char *getDiscTitle(int);
unsigned char *getTrackTitle(int, int);
unsigned char *_getTrackInfo(int, int, int, int *);
int *getTrackLength(int);
int BCD2int(unsigned char);
unsigned char *getTrackEncoding(int);
unsigned char *getTrackFlags(int);
void getTrackDateTime(int);



//static Devid *isKnownDevice(int, int);
//libusb_device **findDevice(libusb_device **, int); //Would not accept **
//void openDevice(libusb_device *, libusb_device_handle **);
void setCurrentDevice(int);
//static int findDevice(libusb_device ** ); //Would not accept **
//static int openDevice();
int mdusb_init();
int mdusb_close();
unsigned char *sendCommand(unsigned char *, int, int *);


/*
unsigned getbits(unsigned x, int p, int n);
int querySize(char *q, va_list args);
char* formatQuery(int *qsize, char *c, ...);
char* scanQuery(int qsize, char *result, char *c, char *f);
int mdInit();
int mdClose();

static int findDevice(libusb_device **devs);
static int openDevice();
int mdusb_init();
int mdusb_close();
char *sendCommand(char *buffer3, char *data, int length, int *rsize);
*/

#endif /*MD_H_*/ 
