#include <stdio.h>
#include <stdlib.h>
#include "md.h"
#include <sys/types.h>
#include <malloc.h>
#include <libusb-1.0/libusb.h>


Devid known_dev[] = { 
 {0x04dd, 0x7202, "Sharp IM-MT899H"},
 {0x054c, 0x0075, "Sony MZ-N1"}, 
 {0x054c, 0x0080, "Sony LAM-1"}, 
 {0x054c, 0x0081, "Sony MDS-JB980"}, 
 {0x054c, 0x0084, "Sony MZ-N505"}, 
 {0x054c, 0x0085, "Sony MZ-S1"}, 
 {0x054c, 0x0086, "Sony MZ-N707"}, 
 {0x054c, 0x00c6, "Sony MZ-N10"}, 
 {0x054c, 0x00c7, "Sony MZ-N910"},
 {0x054c, 0x00c8, "Sony MZ-N710/NF810"}, 
 {0x054c, 0x00c9, "Sony MZ-N510/N610"}, 
 {0x054c, 0x00ca, "Sony MZ-NE410/NF520D"}, 
 {0x054c, 0x00eb, "Sony MZ-NE810/NE910"},
 {0x054c, 0x0101, "Sony LAM-10"},
 {0x054c, 0x0113, "Aiwa AM-NX1"},
 {0x054c, 0x014c, "Aiwa AM-NX9"},
 {0x054c, 0x017e, "Sony MZ-NH1"},
 {0x054c, 0x0180, "Sony MZ-NH3D"},
 {0x054c, 0x0182, "Sony MZ-NH900"},
 {0x054c, 0x0184, "Sony MZ-NH700/NH800"},
 {0x054c, 0x0186, "Sony MZ-NH600/NH600D"},
 {0x054c, 0x0188, "Sony MZ-N920"},
 {0x054c, 0x018a, "Sony LAM-3"},
 {0x054c, 0x01e9, "Sony MZ-DH10P"},
 {0x054c, 0x0219, "Sony MZ-RH10"},
 {0x054c, 0x021b, "Sony MZ-RH710/MZ-RH910"},
 {0x054c, 0x022c, "Sony CMT-AH10 (stereo set with integrated MD)"},
 {0x054c, 0x023c, "Sony DS-HMD1 (device without analog music rec/playback)"},
 {0x054c, 0x0286, "Sony MZ-RH1"}
};

//			unsigned char serial[SERIAL_LENGTH];
//const int MD_SERIALNO_LENGTH = 256; //FIXME: Are serial numbers guarantied not to exceed 256?
char serial[MD_SERIALNO_LENGTH];
libusb_device_handle *cur_handle = NULL; //The device to which commands are sent
libusb_context *ctx; //Used by mdusb_init() and mdusb_close()
MDdevice *mds; //List of minidisc devices found
int cntmd; //Number of minidisc devices found

static Devid *isKnownDevice(int idVendor, int idProduct){
	int i = 0;
	for(i = 0; i < sizeof(known_dev) / sizeof(Devid); i++){
		if(known_dev[i].idVendor == idVendor && known_dev[i].idProduct == idProduct){
			return &known_dev[i]; //FIXME: Should be ok, but check
		}
	}
	return NULL;
}

/*
* **devs - list of devices attaced to the system
*	cnt - number of devices in the list
*
*	Loop through the attatched devices, picking out the mds, then place
*	the found mds in nonlocal array of struct mds
*	If a device_descriptor of an attached device cannot be fetched the
*	function will return an error
*
*
*	Malloc:
*					devices
*					mds
*/
//static int findDevice(libusb_device **devs)
static int findDevice(libusb_device **devs, libusb_device ***founddevs, int cnt)
{	
	cntmd = 0;
	libusb_device *tmpdev;
	libusb_device **devices;
	//Freed in mdusb_init
	devices = malloc(sizeof(void *) * cnt + 1);//NULL terminated
	mds = malloc(sizeof(MDdevice) * cnt); //Freed in mdusb_close()
	int i = 0;
	int j = 0;
	while ((tmpdev = devs[i++]) != NULL) {
		struct libusb_device_descriptor desc;
		int r = libusb_get_device_descriptor(tmpdev, &desc);
		if (r < 0) {
			DBG("%s\n", "Failed to get device descriptor");
			return r;
		}

		Devid *mddev;		
		if ((mddev = isKnownDevice(desc.idVendor, desc.idProduct))){
				devices[j] = tmpdev;
			cntmd++;  //Used in mdclose()
			mds[j].idVendor = desc.idVendor;
			mds[j].idProduct = desc.idProduct;
			mds[j++].description = mddev->description;
			//FIXME: Add device serial number
			DBG("Found a %s %04x:%04x (bus %d, device %d)\n",
				mddev->description, desc.idVendor, desc.idProduct,
				libusb_get_bus_number(tmpdev), libusb_get_device_address(tmpdev));
		}
	}
	devices[j] = NULL;
	*founddevs = devices;
	return 0;
}


//data - data to send
//length - length of data to send
//rsize - set to size of reply
//char *testCommand(libusb_device_handle *handle){
unsigned char *sendCommand(unsigned char *data, int length, int *rsize){
	//FIXME: Use lib usb masks
	int r = 0;
	int i = 0;

	i = 0;
	DBG("%s", "USB: Sending the following query:\n");
	for(i = 0; i < length; i++){
		DBGPLAIN("%02x ", data[i]);
	}
	DBGPLAIN("\n");

	//Send query
	r = libusb_control_transfer(cur_handle, 0x41, 0x80, 0, 0, data, length, 0);
	DBG("USB: Number of byte transferred in first call: %i\n", r);

	unsigned char *buffer2;
	buffer2 = malloc(4); 
	i = 0;
	for(i = 0; i < 4; i++){
		buffer2[i] = 0;
	}

	//Get size of reply
	DBG("%s\n", "USB: Fetch size of reply:");
	r = libusb_control_transfer(cur_handle, 0xC1, 0x01, 0, 0, buffer2, 4, 0);
	DBG("USB: Number of byte transferred in second call: %i\n", r);
	DBGPLAIN("Printing size array: \n");
	i = 0;
	for(i = 0; i < 4; i++){
		DBGPLAIN("Slot nos: %i value %i\n", i, buffer2[i]);
	}

	uint16_t rlength;
	//Create buffer to hold result
	rlength = (uint16_t) buffer2[2];
	unsigned char *buffer3;
	buffer3 = malloc(buffer2[2]); 
	i = 0;
	for(i = 0; i < buffer2[2]; i++){
		buffer3[i] = 0;
	}

	//Fetch reply
	DBG("%s\n", "USB: Fetch reply:");
	r = libusb_control_transfer(cur_handle, 0xC1, 0x81, 0, 0, buffer3, rlength, 0);
	DBG("Number of byte transferred in third call: %i\n", r);

	//FIXME: Add proper error handling
	//Caller must check for errors prior to calling scanQuery
	if(buffer3[0] != STATUS_ACCEPTED){
		fprintf(stderr, "STATUS_ERROR_OCCURED: %x\n", buffer3[0]);
	}else{
		DBG("USB: STATUS_ACCEPTED: %x\n", buffer3[0]);
	}

	unsigned char *result;
	result = malloc(buffer2[2] - 1);

	i = 0;
	DBG("%s", "USB: Reply received:\n");
	for(i = 0; i < buffer2[2] - 1; i++){
		result[i] = buffer3[i+1];
		DBGPLAIN("%02x ", result[i]);
	}
	DBGPLAIN("\n");

	//Set rsize - informs caller of the size of the reply
	*rsize = buffer2[2] - 1;

	free(buffer2);
	free(buffer3);

	return result;
}

/*
static int setSerialNo(MDdevice mddev, libusb_device *dev){
	struct libusb_device_descriptor desc;
	int r = libusb_get_device_descriptor(dev, &desc);
	if (r < 0) {
		fprintf(stderr, "%s\n", "Failed to get device descriptor");
		return r;
	}
	r = 0;
	if(!(r = libusb_get_string_descriptor_ascii(mddev.handle, desc.iSerialNumber,
		 mddev.serialno, MD_SERIALNO_LENGTH))){
		fprintf(stderr, "%s\n", "Failed to get device descriptor");
		return r;
	}
	return 0;
}
*/

int mdusb_init()
{
	int r;
	ssize_t cnt;
	libusb_device **devs;
	libusb_device **founddevs;
	//Initialise libusb
	r = libusb_init(&ctx);
	if (r < 0)
		return r;

	libusb_set_debug(ctx, 3); 

	cnt = libusb_get_device_list(ctx, &devs);
	if (cnt < 0)
		return (int) cnt;

	r = findDevice(devs, &founddevs, cnt);
	if(r){
		DBG("The device descriptor for a device coudld not be retrieved. r = %d\n", r);
		return 1;
	}

	//FIXME: Modify later
	if(founddevs[0] == NULL){
		fprintf(stderr, "No minidisc devices were found");
	  return 1;
	}

	libusb_device_handle *handle;
	libusb_device *tmpdev;
	int i = 0;
	while ((tmpdev = founddevs[i]) != NULL) {
		DBG("%s\n", "Trying to open device.");
		r = libusb_open(tmpdev, &handle);
		if(r){
			fprintf(stderr, "Failed to open device. Error %d\n", r); 
			return r;
		}
		else{
			mds[i].handle = handle;
//			if(setSerialNo(mds[i], tmpdev)){
//				DBG("%s", "Unable to retrive device serial number");
//			};
		}
	i++;
	}

	i = 0;
	DBG("%s", "The following devices were opened:\n");
	while(i < cntmd){
		DBGPLAIN("%s: %04x:%04x\n", mds[i].description, mds[i].idVendor, mds[i].idProduct);
		i++;
	}

//Start: For demo purposes
	i = 0;
	printf("%s", "The following devices were opened:\n");
	while(i < cntmd){
		printf("%s: %04x:%04x\n", mds[i].description, mds[i].idVendor, mds[i].idProduct);
		i++;
	}
	printf("\n");
//End

	free(founddevs);
	libusb_free_device_list(devs, 1);
	setCurrentDevice(0);
	return 0;
}


void setCurrentDevice(int devno){
	cur_handle = mds[devno].handle;
} 


int mdusb_close(){
	DBG("%s", "Enter mdusb_close()\n");
	int i = 0;
	while(i < cntmd){
		libusb_close(mds[i++].handle);
	}
	free(mds);
	libusb_exit(ctx);
	return 0;
}
